# Copyright 2025-2026  Dom Sekotill <dom.sekotill@kodo.org.uk>

"""
Curl handle configuration supporting various TLS backends
"""

from __future__ import annotations

import hashlib
import re
from os import fspath
from pathlib import Path
from tempfile import mkdtemp
from typing import TypeAlias
from typing import TypeVar
from typing import assert_never
from typing import overload

import pycurl

from ..abc import ConfigHandle
from .detect import identify_certificate_file
from .encodings import AsciiArmored
from .encodings import Certificate
from .encodings import Encoding
from .encodings import Pkcs12
from .encodings import PrivateKey
from .files import EncodedFile

ContainerT = TypeVar("ContainerT", AsciiArmored, Pkcs12)
EncodedT = TypeVar("EncodedT", bound=Encoding)
RawT = TypeVar("RawT", Certificate, PrivateKey)

CommonEncodedSource: TypeAlias = (
	AsciiArmored | Pkcs12 | EncodedFile[AsciiArmored] | EncodedFile[Pkcs12]
)
EncodedSource: TypeAlias = CommonEncodedSource | RawT | EncodedFile[RawT]

CertificateSource: TypeAlias = EncodedSource[Certificate]
PrivateKeySource: TypeAlias = EncodedSource[PrivateKey]

__all__ = [
	"CertificateSource",
	"PrivateKeySource",
	"add_ca_certificate",
	"add_client_certificate",
]

_tempdir: Path | None = None


def temp_dir() -> Path:
	"""
	Idempotently create and return a temporary directory
	"""
	global _tempdir
	if _tempdir is None:
		_tempdir = Path(mkdtemp())
	return _tempdir


def add_ca_certificate(  # noqa: C901
	handle: ConfigHandle,
	cert_source: CertificateSource | Path,
) -> None:
	"""
	Configure a handle with Certificate Authority certificates
	"""
	if isinstance(cert_source, Path):
		if cert_source.is_dir():
			handle.setopt(pycurl.CAPATH, fspath(cert_source))
			return
		cert_source = identify_certificate_file(cert_source)

	match pycurl.version_info()[5].lower().split("/"):
		case ["gnutls", _]:
			use_blob = False
		case _:
			use_blob = True

	match cert_source:
		case EncodedFile() if isinstance(cert_source.contents, AsciiArmored):
			handle.setopt(pycurl.CAINFO, fspath(cert_source.path))
		case EncodedFile() if use_blob:
			if not (cert := cert_source.contents.certificate()):
				msg = f"no certificate found in {cert_source!r}"
				raise ValueError(msg)
			handle.setopt(pycurl.CAINFO_BLOB, AsciiArmored.new(cert))
		case EncodedFile():
			if not (cert := cert_source.contents.certificate()):
				msg = f"no certificate found in {cert_source!r}"
				raise ValueError(msg)
			cert_source = _container_file(AsciiArmored, cert, None)
			handle.setopt(pycurl.CAINFO, fspath(cert_source.path))
		case AsciiArmored() if use_blob:
			handle.setopt(pycurl.CAINFO_BLOB, cert_source.to_bytes())
		case AsciiArmored():
			cert_source = _as_file(cert_source)
			handle.setopt(pycurl.CAINFO, fspath(cert_source.path))
		case _:
			if not (cert := cert_source.certificate()):
				msg = f"no certificate found in {cert_source!r}"
				raise ValueError(msg)
			cert_source = _container_file(AsciiArmored, cert, None)
			handle.setopt(pycurl.CAINFO, fspath(cert_source.path))


@overload
def add_client_certificate(
	handle: ConfigHandle,
	cert: CertificateSource,
	key: PrivateKeySource,
) -> None: ...


@overload
def add_client_certificate(
	handle: ConfigHandle,
	cert: CommonEncodedSource,
	key: None = None,
) -> None: ...


def add_client_certificate(
	handle: ConfigHandle,
	cert: CertificateSource,
	key: PrivateKeySource | None = None,
) -> None:
	"""
	Configure a handle with a client certificate
	"""
	match pycurl.version_info()[5].lower().split("/"):
		case ["openssl", str(version)]:
			_configure_openssl_handle(handle, version, cert, key)
		case ["gnutls", str(version)]:
			_configure_gnutls_handle(handle, version, cert, key)
		case ["mbedtls", _]:
			_configure_mbedtls_handle(handle, cert, key)
		case ["wolfssl", _]:
			_configure_wolfssl_handle(handle, cert, key)
		case ["schannel" | "secure transport", _]:
			_configure_pkcs12_handle(handle, cert, key)
		case _:
			# TODO(dom): Get samples of other backends' version strings
			# https://code.kodo.org.uk/konnect/konnect.curl/-/issues/10
			raise NameError("unknown TLS backend")


def _configure_openssl_handle(
	handle: ConfigHandle,
	version_str: str,
	cert: CertificateSource,
	key: PrivateKeySource | None,
) -> None:
	# OpenSSL >= 0.9.3 supports PKCS#12; all formats as files or PEM and PKCS#12 as blobs
	# Keys can be given as a separate blob with CURLOPT_SSLKEY_BLOB
	version = _split_version(version_str)
	match cert:
		case Certificate():
			handle.setopt(pycurl.SSLCERTTYPE, AsciiArmored.format)
			handle.setopt(pycurl.SSLCERT_BLOB, AsciiArmored.new(certificate=cert))
		case Pkcs12() if version < [0, 9, 3]:
			cert = _container_blob(AsciiArmored, cert, None)
			handle.setopt(pycurl.SSLCERTTYPE, cert.format)
			handle.setopt(pycurl.SSLCERT_BLOB, cert)
		case AsciiArmored() | Pkcs12():
			handle.setopt(pycurl.SSLCERTTYPE, cert.format)
			handle.setopt(pycurl.SSLCERT_BLOB, cert)
		case EncodedFile():
			handle.setopt(pycurl.SSLCERTTYPE, cert.contents.format)
			handle.setopt(pycurl.SSLCERT, fspath(cert.path))
		case _ as never:
			assert_never(never)

	match key:
		case None:
			return
		case PrivateKey():
			handle.setopt(pycurl.SSLKEYTYPE, AsciiArmored.format)
			handle.setopt(pycurl.SSLKEY_BLOB, AsciiArmored.new(private_key=key))
		case Pkcs12() if version < [0, 9, 3]:
			cert = _container_blob(AsciiArmored, cert, key)
			handle.setopt(pycurl.SSLCERTTYPE, cert.format)
			handle.setopt(pycurl.SSLCERT_BLOB, cert)
		case AsciiArmored() | Pkcs12():
			handle.setopt(pycurl.SSLKEYTYPE, key.format)
			handle.setopt(pycurl.SSLKEY_BLOB, key)
		case EncodedFile():
			handle.setopt(pycurl.SSLKEYTYPE, key.contents.format)
			handle.setopt(pycurl.SSLKEY, fspath(key.path))
		case _ as never:
			assert_never(never)


def _configure_gnutls_handle(
	handle: ConfigHandle,
	version_str: str,
	cert: CertificateSource,
	key: PrivateKeySource | None,
) -> None:
	# From GnuTLS >= 8.11.0 PEM and P12 are allowed, just PEM before that.
	# Only CURLOPT_SSLCERT works with GnuTLS
	version = _split_version(version_str)

	match cert:
		case _ if key is not None:
			cert = _container_file(AsciiArmored, cert, key)
		case AsciiArmored():
			cert = _as_file(cert)
		case Pkcs12():
			cert = _as_file(cert)
		case EncodedFile() if isinstance(cert.contents, Pkcs12) and version < [8, 11, 0]:
			cert = _container_file(AsciiArmored, cert, None)
		case EncodedFile() if isinstance(cert.contents, AsciiArmored | Pkcs12):
			pass
		case _:
			cert = _container_file(AsciiArmored, cert, None)

	assert isinstance(cert, EncodedFile)
	handle.setopt(pycurl.SSLCERTTYPE, cert.contents.format)
	handle.setopt(pycurl.SSLCERT, fspath(cert.path))


def _configure_mbedtls_handle(
	handle: ConfigHandle,
	cert: CertificateSource,
	key: PrivateKeySource | None,
) -> None:
	# MbedTLS supports PEM and DER encodings with both CURLOPT_SSLCERT and
	# CURLOPT_SSLCERT_BLOB, and only PEM (?) with CURLOPT_SSLKEY. CURLOPT_SSLKEY_BLOB does
	# not appear to be supported.
	match cert:
		case Certificate() | AsciiArmored():
			handle.setopt(pycurl.SSLCERTTYPE, cert.format)
			handle.setopt(pycurl.SSLCERT_BLOB, cert)
		case EncodedFile() if not isinstance(cert.contents, Pkcs12):
			handle.setopt(pycurl.SSLCERTTYPE, cert.contents.format)
			handle.setopt(pycurl.SSLCERT, fspath(cert.path))
		case _:
			blob = _container_blob(AsciiArmored, cert, key)
			handle.setopt(pycurl.SSLCERTTYPE, blob.format)
			handle.setopt(pycurl.SSLCERT_BLOB, blob)
			return

	match key:
		case None:
			return
		case EncodedFile() if isinstance(key.contents, AsciiArmored):
			handle.setopt(pycurl.SSLKEYTYPE, "PEM")
			handle.setopt(pycurl.SSLKEY, fspath(key.path))
			return
		case EncodedFile():
			key = key.contents.private_key()
		case _:
			key = key.private_key()

	file = _as_file(AsciiArmored.new(private_key=key))
	handle.setopt(pycurl.SSLKEYTYPE, "PEM")
	handle.setopt(pycurl.SSLKEY, fspath(file.path))


def _configure_wolfssl_handle(
	handle: ConfigHandle,
	cert: CertificateSource,
	key: PrivateKeySource | None,
) -> None:
	# WolfSSL is similar to mbedtls, but supports CURLOPT_SSLKEY_BLOB
	match cert:
		case Certificate() | AsciiArmored():
			handle.setopt(pycurl.SSLCERTTYPE, cert.format)
			handle.setopt(pycurl.SSLCERT_BLOB, cert)
		case EncodedFile() if not isinstance(cert.contents, Pkcs12):
			handle.setopt(pycurl.SSLCERTTYPE, cert.contents.format)
			handle.setopt(pycurl.SSLCERT, fspath(cert.path))
		case _:
			blob = _container_blob(AsciiArmored, cert, key)
			handle.setopt(pycurl.SSLCERTTYPE, blob.format)
			handle.setopt(pycurl.SSLCERT_BLOB, blob)
			return

	match key:
		case None:
			return
		case PrivateKey() | AsciiArmored():
			pass
		case EncodedFile() if isinstance(key.contents, AsciiArmored):
			handle.setopt(pycurl.SSLKEYTYPE, key.contents.format)
			handle.setopt(pycurl.SSLKEY, fspath(key.path))
			return
		case EncodedFile():
			key = key.contents.private_key()
		case _:
			key = key.private_key()

	if key is None:
		msg = f"no private key found in {key}"
		raise ValueError(msg)

	handle.setopt(pycurl.SSLKEYTYPE, key.format)
	handle.setopt(pycurl.SSLKEY_BLOB, key)


def _configure_pkcs12_handle(
	handle: ConfigHandle,
	cert: CertificateSource,
	key: PrivateKeySource | None,
) -> None:
	# Both Schannel and Secure Transport supports only PKCS#12, for both file and blob types
	match cert:
		case EncodedFile() as file if isinstance(file.contents, Pkcs12) and key is None:
			handle.setopt(pycurl.SSLCERTTYPE, file.contents.format)
			handle.setopt(pycurl.SSLCERT, fspath(file.path))
			return
		case Pkcs12() as blob if key is None:
			handle.setopt(pycurl.SSLCERTTYPE, blob.format)
			handle.setopt(pycurl.SSLCERT_BLOB, blob)
			return
		case _:
			blob = _container_blob(Pkcs12, cert, key)
			handle.setopt(pycurl.SSLCERTTYPE, blob.format)
			handle.setopt(pycurl.SSLCERT_BLOB, blob)


def _container_blob(
	cls: type[ContainerT],
	cert_source: CertificateSource,
	key_source: PrivateKeySource | None,
) -> ContainerT:
	match cert_source:
		case AsciiArmored() | Pkcs12():
			cert = cert_source.certificate()
			key = cert_source.private_key()
		case EncodedFile():
			cert = cert_source.contents.certificate()
			key = cert_source.contents.private_key()
		case Certificate() as cert:
			key = None
		case _ as never:
			assert_never(never)

	match key_source:
		case None:
			pass
		case AsciiArmored() | Pkcs12():
			key = key_source.private_key()
		case EncodedFile():
			key = key_source.contents.private_key()
		case PrivateKey() as key:
			pass
		case _ as never:
			assert_never(never)

	if cert is None:
		msg = f"no certificate found in {cert_source!r}"
		raise ValueError(msg)
	if key is None:
		msg = f"no private key found in {key_source or cert_source!r}"
		raise ValueError(msg)

	return cls.new(certificate=cert, private_key=key)


def _container_file(
	cls: type[ContainerT],
	cert_source: CertificateSource,
	key_source: PrivateKeySource | None,
) -> EncodedFile[ContainerT]:
	blob = _container_blob(cls, cert_source, key_source)
	return _as_file(blob)


def _as_file(encoded_data: EncodedT) -> EncodedFile[EncodedT]:
	sha1 = hashlib.sha1()
	if cert := encoded_data.certificate():
		sha1.update(cert)
	if key := encoded_data.private_key():
		sha1.update(key)
	path = temp_dir() / sha1.hexdigest()

	try:
		return EncodedFile.write(path, encoded_data)
	except FileExistsError:
		# Assumes if the file exists this data has already been written to it
		return EncodedFile(encoded_data, path)


def _split_version(version: str) -> list[int]:
	return [int(part) for part in re.findall(r"(?:^|(?<=[._-]))[0-9]+", version)]



# Copyright 2025-2026  Dom Sekotill <dom.sekotill@kodo.org.uk>

"""
File storage helper class for encoded cryptographic formats
"""

from __future__ import annotations

from pathlib import Path
from typing import Final
from typing import Generic
from typing import TypeVar
from typing import final

from .encodings import Encoding

T = TypeVar("T", bound=Encoding, covariant=True)
C = TypeVar("C", bound=Encoding)

DEFAULT_MAX_READ: Final = 2**16  # 64kiB


@final
class EncodedFile(Generic[T]):
	"""
	Combines decoding data with file path information
	"""

	def __init__(self, contents: T, path: Path) -> None:
		self.contents = contents
		self.path = path

	@classmethod
	def read(
		cls: type[EncodedFile[C]],
		path: Path,
		encoding: type[C],
		/,
		maxsize: int = DEFAULT_MAX_READ,
	) -> EncodedFile[C]:
		"""
		Read encoded data from the file path if it exists and return an `EncodedFile`

		The value of 'maxsize' is a safety net to prevent arbitrarily large files being read
		into memory.  The default should be suitable for most certificate and key files but
		can be overridden to allow unusually large files to be read (probably ASCII armored
		files containing many items).

		Can raise the normal range of `OSError` exceptions that may occur when opening and
		reading a file.
		"""
		with path.open("rb") as handle:
			contents = encoding.from_bytes(handle.read(maxsize))
		return cls(contents, path)

	@staticmethod
	def write(path: Path, contents: C, /, *, exists_ok: bool = False) -> EncodedFile[C]:
		"""
		Write encoded data to the file path

		If 'exists_ok' is false the file will be opened in create-only mode, raising
		`FileExistsError` if there is already a file at the path; otherwise the file is
		opened in normal write mode and truncated before writing the encoded data.  Either
		way the data is never appended to the file.

		Can raise the normal range of `OSError` exceptions that may occur when opening and
		writing to a file.
		"""
		# Its a shame this cannot be a @classmethod to inject EncodedFile subclasses but it
		# cannot be done without subscripting the subclass with the contents type when
		# called.
		mode = "wb" if exists_ok else "xb"
		with path.open(mode) as handle:
			handle.write(contents.to_bytes())
		return EncodedFile(contents, path)



# Copyright 2023-2024  Dom Sekotill <dom.sekotill@kodo.org.uk>

"""
Abstract protocols which may be implemented by users

`konnect.curl` provides simple implementations of these protocols which may be used directly
or subclassed by users.
"""

from typing import Protocol
from typing import TypeVar

U_co = TypeVar("U_co", covariant=True)
R_co = TypeVar("R_co", covariant=True)


# TODO(dom): Make concrete proxies and convert Pycurl exceptions to package exceptions
# https://code.kodo.org.uk/konnect/konnect.curl/-/work_items/6


class ConfigHandle(Protocol):
	"""
	The interface provided by objects passed to `RequestProtocol.configure_handle()`
	"""

	def setopt(self, option: int, value: object, /) -> None:
		"""
		Set an option on a Curl handle

		See https://curl.se/libcurl/c/curl_easy_setopt.html for a list of options and what
		they do.
		"""
		...

	def unsetopt(self, option: int, /) -> None:
		"""
		Set options that have a default value back to that default
		"""
		...

	def pause(self, state: int, /) -> None:
		"""
		Set the paused state of a Curl handle

		This is provided for request implementations that need to await upload data, they
		may store the handle or method for later use. In a future release it will be
		deprecated in favour of a method that returns an asynchronously writable object.
		"""
		...


class GetInfoHandle(Protocol):
	"""
	The interface provided by objects passed to `RequestProtocol.completed()`
	"""

	def getinfo(self, option: int, /) -> object:
		"""
		Return information about a Curl handle

		Note that string values are returned as unicode strings.

		See https://curl.se/libcurl/c/curl_easy_getinfo.html for a list of options and what
		they return.
		"""
		...

	def getinfo_raw(self, option: int, /) -> object:
		"""
		Like `getinfo()` but string values are returned as byte strings
		"""
		...


class RequestProtocol(Protocol[U_co, R_co]):
	"""
	Request classes that are passed to `Multi.process()` must implement this protocol
	"""

	def configure_handle(self, handle: ConfigHandle, /) -> None:
		"""
		Configure a Curl handle for the request by calling `ConfigHandle` methods

		See https://curl.se/libcurl/c/curl_easy_setopt.html for a list of options and what
		they do.
		"""
		...

	def has_update(self) -> bool:
		"""
		Return whether calling `get_update()` will return a value or raise `LookupError`
		"""
		...

	def get_update(self) -> U_co:
		"""
		Return a waiting update or raise `LookupError` if there is none

		`Multi.process()` will only call this method when `has_update()` indicates an update
		is available.  The returned value will be returned by `Multi.process()`.

		Note that values returned by this method are interim updates, and `Multi.process()`
		will be called again with the current request.  It is up to the implementer how many
		times updates will be returned and what objects to return as updates: it may be
		different objects for different stages of a transfer; or there may never be interim
		updates.
		"""
		...

	def completed(self, handle: GetInfoHandle, /) -> R_co:
		"""
		Indicate that Curl has completed processing the handle and return a final response

		Like `get_update` this method's return value will be returned by `Multi.process()`.
		Unlike `get_update` this method will be called exactly once for a successful
		transfer.

		The `GetInfoHandle` passed as a positional argument may be used to get
		post-completion information about a transfer, see
		https://curl.se/libcurl/c/curl_easy_getinfo.html for a list of options and what they
		return.
		"""
		...



# Copyright 2025  Dom Sekotill <dom.sekotill@kodo.org.uk>

"""
Tests for the konnect.curl.certificates.ascii_armor module
"""

import unittest
from hashlib import sha256
from importlib import resources
from io import BytesIO
from textwrap import dedent
from typing import assert_never

from konnect.curl.certificates.ascii_armor import ArmoredData


class Tests(unittest.TestCase):
	"""
	Tests for the AsciiArmored class
	"""

	def test_encode_lines(self) -> None:
		"""
		Check encoding ArmoredData produces the expected output
		"""
		raw = """
		And the Lord spake, saying, "First shalt thou take out the Holy Pin. Then shalt thou
		count to three, no more, no less. Three shall be the number thou shalt count, and
		the number of the counting shall be three. Four shalt thou not count, neither count
		thou two, excepting that thou then proceed to three. Five is right out. Once the
		number three, being the third number, be reached, then lobbest thou thy Holy Hand
		Grenade of Antioch towards thy foe, who, being naughty in My sight, shall snuff it.
		"""
		data = ArmoredData(
			"TEST", dedent(raw).lstrip().encode("ascii"), [("Ref", "Arm 2:9-21"), ("Lang", "Latin")]
		)
		digest = "9700a89facf5ae58b6d2d599dd1fce2dcc167d493c17ce8d4b8ad7c218266ee3"

		assert sha256(b"".join(data.encode_lines())).hexdigest() == digest

	def test_extract(self) -> None:
		"""
		Check that data is extracted accurately from a sample
		"""
		sample = resources.files("tests") / "sample.pem"
		data = [*ArmoredData.extract(sample.read_bytes())]

		assert data[0].label == "SAMPLE"
		assert data[0].headers == [("Encoding", "base64"), ("Content-Type", "text/plain")]
		assert b"Strange women lying in ponds" in data[0]
		assert b"farcical aquatic ceremony" in data[0]

		assert data[1].label == "PRIVATE KEY"
		assert data[1].headers == []

		assert data[2].label == "CERTIFICATE"
		assert data[2].headers == []

	def test_non_normative_extract(self) -> None:
		"""
		Check that the parser accepts a checksum and is a forgiving of common misformats
		"""
		sample = resources.files("tests") / "non-normative-sample.pem"
		data = [*ArmoredData.extract(sample.read_bytes())]

		assert data[0].label == "SAMPLE"
		assert data[0].headers == [("Encoding", "base64"), ("Content-Type", "text/plain")]
		assert b"Strange women lying in ponds" in data[0]
		assert b"farcical aquatic ceremony" in data[0]

	def test_extract_with_unarmored(self) -> None:
		"""
		Check that unarmored text is extracted verbatim as Unicode strings when asked
		"""
		sample = resources.files("tests") / "sample.pem"
		buff = BytesIO()

		for data in ArmoredData.extract((raw := sample.read_bytes()), with_unarmored=True):
			match data:
				case ArmoredData():
					buff.writelines(data.encode_lines())
				case str():
					buff.write(data.encode("utf-8"))
				case _ as never:
					assert_never(never)

		buff.seek(0)
		assert buff.read() == raw

	def test_extract_incomplete(self) -> None:
		"""
		Check that incomplete or unexpected input raises ValueError
		"""
		lines = [
			b"-----BEGIN SAMPLE-----\n",
			b"Header: Value\n",
			b"\n",
			b"aaa=\n",
			b"---",
			b"--END TEST-----\n",
		]
		sample = bytearray()

		for line in lines:
			sample.extend(line)

			with self.assertRaises(ValueError):
				_ = [*ArmoredData.extract(sample)]



# Copyright 2025  Dom Sekotill <dom.sekotill@kodo.org.uk>

"""
ASCII armor is a 7-bit safe encoding method for various types of data to be embedded in text
"""

import re
from binascii import a2b_base64
from binascii import b2a_base64
from collections.abc import Iterator
from collections.abc import Sequence
from typing import Literal
from typing import Self
from typing import SupportsIndex
from typing import TypeAlias
from typing import overload

BytesType: TypeAlias = Sequence[SupportsIndex]
Header: TypeAlias = tuple[str, str]

CHR_EQL = ord("=")
CHR_DASH = ord("-")


class ArmoredData(bytes):
	"""
	Armored data is a labeled binary data object with optional headers

	Common examples of armored data are X.509 certificates and private keys labeled as
	CERTIFICATE or PRIVATE KEY respectively.
	"""

	def __new__(  # noqa: D102
		cls, label: str, source: BytesType, headers: Sequence[Header] | None = None
	) -> Self:
		return super().__new__(cls, source)

	def __init__(
		self, label: str, source: BytesType, headers: Sequence[Header] | None = None
	) -> None:
		self.label = label
		self.headers: Sequence[Header] = [] if headers is None else list(headers)

	def encode_lines(self) -> Iterator[bytes]:
		"""
		Encode an `ArmoredData` instance and return a line iterator

		The returned iterator yields the encoded field as LF terminated lines and is
		suitable for passing to `typing.IO[bytes].writelines()` or `bytes.join()`.
		"""
		yield f"-----BEGIN {self.label}-----\n".encode("ascii")

		for header, value in self.headers:
			yield f"{header}: {value}\n".encode("ascii")
		if self.headers:
			yield b"\n"

		view = memoryview(self)
		for offset in range(0, len(self), 48):
			yield b2a_base64(view[offset : offset + 48])

		yield f"-----END {self.label}-----\n".encode("ascii")

	@overload
	@classmethod
	def extract(
		cls, data: bytes | bytearray, *, with_unarmored: Literal[True]
	) -> Iterator[Self | str]: ...

	@overload
	@classmethod
	def extract(
		cls, data: bytes | bytearray, *, with_unarmored: Literal[False] = False
	) -> Iterator[Self]: ...

	@classmethod
	def extract(  # noqa: C901
		cls, data: bytes | bytearray, *, with_unarmored: bool = False
	) -> Iterator[Self | str]:
		"""
		Scan input and return an iterator that yields `ArmoredData` and unarmored lines

		An unarmored line is anything outside of any field's start and end lines and is
		yielded as Unicode strings if 'with_unarmored' is `True`.  Note that unarmored lines
		are assumed to be ASCII or UTF-8 encoded.
		"""
		begin_pattern = re.compile(rb"^-----[ ]?BEGIN ([^-]+)-----$")
		end_pattern = re.compile(rb"^-----[ ]?END ([^-]+)-----$")

		headers = list[tuple[str, str]]()

		lines = enumerate(data.splitlines(keepends=True))
		while 1:
			del headers[:]

			for _, line in lines:
				if match := begin_pattern.match(line):
					break
				if with_unarmored:
					yield line.decode("utf-8")
			else:
				return
			label = match.group(1).rstrip()

			for _, line in lines:
				name, is_hdr, val = line.decode("ascii").partition(":")
				if not is_hdr:
					break
				headers.append((name, val.strip()))
			else:
				raise ValueError("missing contents: expecting encoded data")

			buff = bytearray(a2b_base64(line))
			for number, line in lines:
				match line[0]:
					case 0x3D:  # "="
						# skip checksums
						continue
					case 0x2D:  # "-"
						# break on hitting what is most likely the END line
						line_no = number
						break
					case _:
						buff.extend(a2b_base64(line))
			else:
				raise ValueError("missing contents: expecting an END line")

			if not (match := end_pattern.match(line)):
				raise ValueError(f"malformed line encountered ({line_no})")
			if match.group(1).rstrip() != label:
				raise ValueError(f"unmatched END line encountered ({line_no})")

			yield cls(label.decode("ascii"), buff, headers)



# Copyright 2025  Dom Sekotill <dom.sekotill@kodo.org.uk>

"""
Classes for various DER encoded objects, with ASCII armoring and file storage for them
"""

from __future__ import annotations

import hashlib
from collections.abc import Iterator
from typing import TYPE_CHECKING
from typing import ClassVar
from typing import Final
from typing import Protocol
from typing import Self
from typing import overload

from .ascii_armor import ArmoredData

if TYPE_CHECKING:
	from typing import TypeVar

	PrivateKeyT = TypeVar("PrivateKeyT", bound=PrivateKey)

__all__ = [
	"AsciiArmored",
	"Certificate",
	"ECPrivateKey",
	"Pkcs8EncryptedPrivateKey",
	"Pkcs8PrivateKey",
	"Pkcs12",
	"PrivateKey",
	"RSAPrivateKey",
]

DEFAULT_MAX_READ: Final = 2**16  # 64kiB


class Encoding(Protocol):
	format: ClassVar[str]

	@classmethod
	def from_bytes(cls, source: bytes, /) -> Self: ...
	def to_bytes(self) -> bytes: ...

	def certificate(self) -> Certificate | None: ...
	def private_key(self) -> PrivateKey | None: ...


class Certificate(bytes):
	"""
	X.509 certificates
	"""

	format: ClassVar = "DER"
	label: ClassVar = "CERTIFICATE"

	@classmethod
	def from_bytes(cls, source: bytes, /) -> Self:
		"""
		Return a new instance from an in-memory bytes string
		"""
		return cls(source)

	def to_bytes(self) -> Self:
		"""
		Return a bytes string representation of an instance (itself, as it subclasses bytes)
		"""
		return self

	def fingerprint(self) -> str:
		"""
		Return the SHA1 hash of the certificate as a hexadecimal string
		"""
		# TODO(dom): Would like to return hashlib.HASH but not currently allowed by typeshed
		# https://code.kodo.org.uk/konnect/konnect.curl/-/issues/9
		return hashlib.sha1(self).hexdigest()

	def certificate(self) -> Self:
		"""
		Return the certificate, itself
		"""
		return self

	def private_key(self) -> None:
		"""
		Return None, this is a no-op for certificates
		"""
		return


class PrivateKey(bytes):
	"""
	Base class for private key containers
	"""

	format: ClassVar[str]
	label: ClassVar[str]

	def __new__(cls, source: bytes, /) -> Self:  # noqa: D102
		if cls is PrivateKey:
			msg = "cannot instantiate PrivateKey base class"
			raise TypeError(msg)
		return bytes.__new__(cls, source)

	def __init_subclass__(cls, label: str) -> None:
		cls.format = "DER"
		cls.label = label

	@classmethod
	def from_bytes(cls, source: bytes, /) -> Self:
		"""
		Return a new instance from an in-memory bytes string
		"""
		return cls(source)

	def to_bytes(self) -> Self:
		"""
		Return a bytes string representation of an instance (itself, as it subclasses bytes)
		"""
		return self

	def certificate(self) -> None:
		"""
		Return None, this is a no-op for private keys
		"""
		return

	def private_key(self) -> Self:
		"""
		Return the private key, itself
		"""
		return self


class RSAPrivateKey(PrivateKey, label="RSA PRIVATE KEY"):
	"""
	PKCS#1 private key
	"""


class Pkcs8PrivateKey(PrivateKey, label="PRIVATE KEY"):
	"""
	PKCS#8 unencrypted private key
	"""


class Pkcs8EncryptedPrivateKey(PrivateKey, label="ENCRYPTED PRIVATE KEY"):
	"""
	PKCS#8 encrypted private key
	"""


class ECPrivateKey(PrivateKey, label="EC PRIVATE KEY"):
	"""
	ECDSA private key
	"""


class AsciiArmored(bytes):
	"""
	Base 64 encoding with fences for binary cryptographic data, commonly known as PEM

	This class assumes that the encoded data contains at most one certificate and one
	private key, the first occurrence of each being returned by the `certificate()` and
	`private_key()` methods respectively.
	"""

	format: ClassVar = "PEM"

	@classmethod
	def new(
		cls, certificate: Certificate | None = None, private_key: PrivateKey | None = None
	) -> Self:
		"""
		Return an instance with the encoded form of the given certificate and/or private key
		"""
		parts = list[bytes]()

		if certificate:
			parts.extend(ArmoredData("CERTIFICATE", certificate).encode_lines())

		if private_key:
			parts.extend(ArmoredData(private_key.label, private_key).encode_lines())

		return cls(b"".join(parts))

	@classmethod
	def from_bytes(cls, source: bytes, /) -> Self:
		"""
		Return a new instance from an in-memory bytes string
		"""
		return cls(source)

	def to_bytes(self) -> Self:
		"""
		Return a bytes string representation of an instance (itself, as it subclasses bytes)
		"""
		return self

	def certificate(self) -> Certificate | None:
		"""
		Return the first certificate found in the encoded data, or None
		"""
		try:
			return self.find_first(Certificate)
		except NameError:
			return None

	def private_key(self) -> PrivateKey | None:
		"""
		Return the first private key found in the encoded data
		"""
		try:
			return self.find_first(PrivateKey)
		except NameError:
			return None

	@overload
	def find_first(self, kind: type[Certificate], /) -> Certificate: ...

	@overload
	def find_first(self, kind: type[PrivateKeyT], /) -> PrivateKeyT: ...

	def find_first(
		self, kind: type[Certificate] | type[PrivateKey]
	) -> Certificate | PrivateKey:
		"""
		Return the first item with a label matching one of the provided types
		"""
		labels = {t.label: t for t in _recurse_subclasses(kind) if t is not PrivateKey}
		for data in ArmoredData.extract(self):
			try:
				cls = labels[data.label]
			except KeyError:
				continue
			assert issubclass(cls, kind)
			return cls(data)
		msg = f"no matching labels found: {str.join(', ', labels)}"
		raise NameError(msg)


class Pkcs12(bytes):
	"""
	An ASN.1 container format for cryptographic data
	"""

	format: ClassVar = "P12"

	@classmethod
	def new(
		cls, certificate: Certificate | None = None, private_key: PrivateKey | None = None
	) -> Self:
		"""
		Return an instance with the encoded form of the given certificate and/or private key
		"""
		raise NotImplementedError

	@classmethod
	def from_bytes(cls, source: bytes, /) -> Self:
		"""
		Return a new instance from an in-memory bytes string
		"""
		return cls(source)

	def to_bytes(self) -> Self:
		"""
		Return a bytes string representation of an instance (itself, as it subclasses bytes)
		"""
		return self

	def certificate(self) -> Certificate | None:
		"""
		Return the first certificate found in the encoded data
		"""
		raise NotImplementedError

	def private_key(self) -> PrivateKey | None:
		"""
		Return the first private key found in the encoded data
		"""
		raise NotImplementedError


def _recurse_subclasses(
	cls: type[Certificate | PrivateKey],
) -> Iterator[type[Certificate | PrivateKey]]:
	yield cls
	for subcls in cls.__subclasses__():
		yield from _recurse_subclasses(subcls)



# Copyright 2025-2026  Dom Sekotill <dom.sekotill@kodo.org.uk>

"""
Functions for detecting encodings of files or in-memory byte strings
"""

from pathlib import Path
from typing import Final

from pyasn1.codec.der import decoder
from pyasn1.error import PyAsn1Error
from pyasn1_modules import rfc5208  # type: ignore
from pyasn1_modules import rfc5280  # type: ignore
from pyasn1_modules import rfc5915  # type: ignore
from pyasn1_modules import rfc7292  # type: ignore
from pyasn1_modules import rfc8017  # type: ignore

from .encodings import AsciiArmored
from .encodings import Certificate
from .encodings import ECPrivateKey
from .encodings import Pkcs8EncryptedPrivateKey
from .encodings import Pkcs8PrivateKey
from .encodings import Pkcs12
from .encodings import PrivateKey
from .encodings import RSAPrivateKey
from .files import EncodedFile

__all__ = [
	"identify_blob",
	"identify_certificate_file",
	"identify_file",
]

MAX_READ_SIZE: Final = 2**14  # 16kiB


def identify_file(
	path: Path,
) -> (
	EncodedFile[AsciiArmored]
	| EncodedFile[Certificate]
	| EncodedFile[Pkcs12]
	| EncodedFile[PrivateKey]
):
	"""
	Return the encoding of a file in the form of an `EncodedFile` instance
	"""
	with path.open("br") as file:
		match identify_blob(file.read(MAX_READ_SIZE)):
			case AsciiArmored() as contents:
				return EncodedFile(contents, path)
			case Certificate() as contents:
				return EncodedFile(contents, path)
			case Pkcs12() as contents:
				return EncodedFile(contents, path)
			case PrivateKey() as contents:
				return EncodedFile(contents, path)


def identify_certificate_file(
	path: Path,
) -> EncodedFile[AsciiArmored] | EncodedFile[Certificate] | EncodedFile[Pkcs12]:
	"""
	Return the encoding of a certificate file in the form of an `EncodedFile` instance
	"""
	with path.open("br") as file:
		match identify_blob(file.read(MAX_READ_SIZE)):
			case AsciiArmored() as contents:
				return EncodedFile(contents, path)
			case Certificate() as contents:
				return EncodedFile(contents, path)
			case Pkcs12() as contents:
				return EncodedFile(contents, path)
			case encoding:
				msg = f"file of type {encoding}, expected a certificate encoding: {path}"
				raise TypeError(msg)


def identify_key_file(
	path: Path,
) -> EncodedFile[AsciiArmored] | EncodedFile[Pkcs12] | EncodedFile[PrivateKey]:
	"""
	Return the encoding of a private key file in the form of an `EncodedFile` instance
	"""
	with path.open("br") as file:
		match identify_blob(file.read(MAX_READ_SIZE)):
			case AsciiArmored() as contents:
				return EncodedFile(contents, path)
			case Pkcs12() as contents:
				return EncodedFile(contents, path)
			case PrivateKey() as contents:
				return EncodedFile(contents, path)
			case encoding:
				msg = f"file of type {encoding}, expected a private key encoding: {path}"
				raise TypeError(msg)


def identify_blob(blob: bytes) -> AsciiArmored | Certificate | Pkcs12 | PrivateKey:
	"""
	Identify the encoding of a blob of octets and return the encoding class

	Raises `ValueError` if the encoding is not one of the supported types.

	Note that ASCII armored data is assumed if the first few octets can be interpreted as an
	ASCII or UTF-8 string.
	"""
	specs = [
		(rfc7292.PFX, Pkcs12),  # PKCS#12
		(rfc5280.Certificate, Certificate),  # X.509
		(rfc8017.RSAPrivateKey, RSAPrivateKey),  # PKCS#1
		(rfc5208.PrivateKeyInfo, Pkcs8PrivateKey),  # PKCS#8
		(rfc5208.EncryptedPrivateKeyInfo, Pkcs8EncryptedPrivateKey),  # PKCS#8
		# (rfc5958.AsymmetricKeyPackage, Pkcs8PrivateKey),  # PKCS#8 v2
		(rfc5915.ECPrivateKey, ECPrivateKey),  # ECDSA
	]
	try:
		blob[:20].decode("utf-8")
	except UnicodeError:
		pass
	else:
		return AsciiArmored(blob)
	for spec, encoding in specs:
		try:
			decoder.decode(blob, asn1Spec=spec())
		except PyAsn1Error:
			continue
		return encoding.from_bytes(blob)
	raise ValueError(f"unable to identify encoding")



# Copyright 2025-2026  Dom Sekotill <dom.sekotill@kodo.org.uk>

"""
Cryptographic certificate management

The support for certificate and key storage formats is very naïve, it does no checks that
keys match certificates.  It exists only to allow users to use as many formats as possible
with as many of libcurl's TLS backends as possible.

The backends currently supported are:
 - OpenSSL
 - GnuTLS
 - MbedSSL
 - WolfSSL
 - Schannel
 - Secure Transport
"""

from .configure import CertificateSource
from .configure import CommonEncodedSource
from .configure import PrivateKeySource
from .configure import add_ca_certificate
from .configure import add_client_certificate
from .encodings import AsciiArmored
from .encodings import Certificate
from .encodings import ECPrivateKey
from .encodings import Pkcs8EncryptedPrivateKey
from .encodings import Pkcs8PrivateKey
from .encodings import Pkcs12
from .encodings import PrivateKey
from .encodings import RSAPrivateKey
from .files import EncodedFile

__all__ = [
	"AsciiArmored",
	"Certificate",
	"CertificateSource",
	"CommonEncodedSource",
	"ECPrivateKey",
	"EncodedFile",
	"Pkcs8EncryptedPrivateKey",
	"Pkcs8PrivateKey",
	"Pkcs12",
	"PrivateKey",
	"PrivateKeySource",
	"RSAPrivateKey",
	"add_ca_certificate",
	"add_client_certificate",
]



# Copyright 2025  Dom Sekotill <dom.sekotill@kodo.org.uk>

"""
Tests for the konnect.curl.certificates.configure.set_client_certificate function
"""

from __future__ import annotations

from collections.abc import Iterator
from importlib import import_module
from importlib import resources
from os import fspath
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Final
from typing import NamedTuple

import pycurl
import pytest

from konnect.curl.certificates import AsciiArmored
from konnect.curl.certificates import Certificate
from konnect.curl.certificates import EncodedFile
from konnect.curl.certificates import Pkcs12
from konnect.curl.certificates import RSAPrivateKey
from konnect.curl.certificates import add_client_certificate

if TYPE_CHECKING:
	from konnect.curl.certificates.configure import CertificateSource
	from konnect.curl.certificates.configure import CommonEncodedSource
	from konnect.curl.certificates.configure import PrivateKeySource

STATIC_SHA1: Final = "da39a3ee5e6b4b0d3255bfef95601890afd80709"


class MockConfigHandle:
	def __init__(self) -> None:
		self.setopt_calls = list[tuple[int, object]]()

	def setopt(self, option: int, value: object) -> None:
		self.setopt_calls.append((option, value))

	def unsetopt(self, option: int) -> None:
		raise NotImplementedError

	def pause(self, state: int) -> None:
		raise NotImplementedError


class MockSHA1:
	def __init__(self) -> None: ...
	def update(self, value: bytes) -> None: ...
	def hexdigest(self) -> str:
		return STATIC_SHA1


class SubtestParamsCombined(NamedTuple):
	msg: str
	version: str
	source: CommonEncodedSource
	opts: list[tuple[int, object]]


class SubtestParamsSeparate(NamedTuple):
	msg: str
	version: str
	cert: CertificateSource
	key: PrivateKeySource
	opts: list[tuple[int, object]]


def test_add_client_certificate(
	subtests: pytest.Subtests, monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
	monkeypatch.setattr(import_module("hashlib"), "sha1", MockSHA1)
	monkeypatch.setattr(
		import_module("konnect.curl.certificates.configure"), "mkdtemp", lambda: tmp_path
	)

	for params in parameter_generator(tmp_path):
		monkeypatch.setattr(
			pycurl, "version_info", lambda version=params.version: [...] * 5 + [version]
		)
		with subtests.test(f"{params.msg} [{params.version}]"):
			if isinstance(params, SubtestParamsCombined):
				try:
					add_client_certificate(handle := MockConfigHandle(), params.source)
				except NotImplementedError:
					if isinstance(params.source, Pkcs12) or (
						isinstance(params.source, EncodedFile) and isinstance(params.source.contents, Pkcs12)
					):
						pytest.xfail("PKCS12 support not fully implemented yet")
					raise
			else:
				try:
					add_client_certificate(handle := MockConfigHandle(), params.cert, params.key)
				except NotImplementedError:
					if (
						isinstance(params.cert, Pkcs12)
						or isinstance(params.key, Pkcs12)
						or (isinstance(params.cert, EncodedFile) and isinstance(params.cert.contents, Pkcs12))
						or (isinstance(params.key, EncodedFile) and isinstance(params.key.contents, Pkcs12))
					):
						pytest.xfail("PKCS12 support not fully implemented yet")
					raise

			assert handle.setopt_calls == params.opts


def parameter_generator(
	tmp_path: Path,
) -> Iterator[SubtestParamsCombined | SubtestParamsSeparate]:
	sample_blob_pem = AsciiArmored(
		resources.files(__name__).joinpath("sample.pem").read_bytes()
	)
	sample_blob_p12 = Pkcs12(resources.files(__name__).joinpath("sample.p12").read_bytes())
	sample_file_x509 = EncodedFile.write(
		tmp_path / "sample.crt", cert := Certificate(b"xxx"), exists_ok=True
	)
	sample_file_rsa = EncodedFile.write(
		tmp_path / "sample.key", key := RSAPrivateKey(b"xxx"), exists_ok=True
	)
	sample_file_pem = EncodedFile.write(
		tmp_path / "sample.pem",
		AsciiArmored.new(certificate=cert, private_key=key),
		exists_ok=True,
	)
	sample_file_p12 = EncodedFile.write(
		tmp_path / "sample.p12",
		Pkcs12(b"xxx"),
		exists_ok=True,
	)

	yield SubtestParamsCombined(
		"combined cert and key (PEM)",
		"OpenSSL/1.1.0",
		sample_blob_pem,
		[
			(pycurl.SSLCERTTYPE, "PEM"),
			(pycurl.SSLCERT_BLOB, sample_blob_pem.to_bytes()),
		],
	)
	yield SubtestParamsCombined(
		"combined cert and key (PKCS12)",
		"OpenSSL/1.1.0",
		sample_blob_p12,
		[
			(pycurl.SSLCERTTYPE, "P12"),
			(pycurl.SSLCERT_BLOB, sample_blob_p12.to_bytes()),
		],
	)
	yield SubtestParamsSeparate(
		"separate cert (PEM) and key (PEM)",
		"OpenSSL/1.1.0",
		sample_blob_pem,
		key := AsciiArmored.new(private_key=RSAPrivateKey(b"xxx")),
		[
			(pycurl.SSLCERTTYPE, "PEM"),
			(pycurl.SSLCERT_BLOB, sample_blob_pem.to_bytes()),
			(pycurl.SSLKEYTYPE, "PEM"),
			(pycurl.SSLKEY_BLOB, key.to_bytes()),
		],
	)
	yield SubtestParamsSeparate(
		"separate cert (PEM) and key (RSA)",
		"OpenSSL/1.1.0",
		sample_blob_pem,
		key := RSAPrivateKey(b"xxx"),
		[
			(pycurl.SSLCERTTYPE, "PEM"),
			(pycurl.SSLCERT_BLOB, sample_blob_pem.to_bytes()),
			(pycurl.SSLKEYTYPE, "PEM"),
			(pycurl.SSLKEY_BLOB, AsciiArmored.new(private_key=key).to_bytes()),
		],
	)
	yield SubtestParamsSeparate(
		"separate cert (x509) and key (PEM)",
		"OpenSSL/1.1.0",
		cert := Certificate(b"xxx"),
		sample_blob_pem,
		[
			(pycurl.SSLCERTTYPE, "PEM"),
			(pycurl.SSLCERT_BLOB, AsciiArmored.new(certificate=cert).to_bytes()),
			(pycurl.SSLKEYTYPE, "PEM"),
			(pycurl.SSLKEY_BLOB, sample_blob_pem.to_bytes()),
		],
	)
	yield SubtestParamsSeparate(
		"PKCS12 cert and RSA key",
		"OpenSSL/0.9.0",
		sample_blob_p12,
		key := RSAPrivateKey(b"xxx"),
		[
			(pycurl.SSLCERTTYPE, "PEM"),
			(pycurl.SSLCERT_BLOB, sample_blob_p12.to_bytes()),
			(pycurl.SSLKEYTYPE, "PEM"),
			(pycurl.SSLKEY_BLOB, AsciiArmored.new(private_key=key).to_bytes()),
		],
	)
	yield SubtestParamsCombined(
		"combined PEM formatted file",
		"OpenSSL/1.1.0",
		sample_file_pem,
		[
			(pycurl.SSLCERTTYPE, "PEM"),
			(pycurl.SSLCERT, fspath(sample_file_pem.path)),
		],
	)
	yield SubtestParamsSeparate(
		"separate cert and key files",
		"OpenSSL/1.1.0",
		sample_file_x509,
		sample_file_rsa,
		[
			(pycurl.SSLCERTTYPE, "DER"),
			(pycurl.SSLCERT, fspath(sample_file_x509.path)),
			(pycurl.SSLKEYTYPE, "DER"),
			(pycurl.SSLKEY, fspath(sample_file_rsa.path)),
		],
	)
	yield SubtestParamsCombined(
		"combined cert and key (PEM)",
		"GnuTLS/8.10.0",
		sample_blob_pem,
		[
			(pycurl.SSLCERTTYPE, "PEM"),
			(pycurl.SSLCERT, fspath(tmp_path / STATIC_SHA1)),
		],
	)
	yield SubtestParamsCombined(
		"combined cert and key (P12)",
		"GnuTLS/8.10.0",
		sample_blob_p12,
		[
			(pycurl.SSLCERTTYPE, "PEM"),
			(pycurl.SSLCERT, fspath(tmp_path / STATIC_SHA1)),
		],
	)
	yield SubtestParamsCombined(
		"combined cert and key (PEM)",
		"GnuTLS/8.11.0",
		sample_blob_pem,
		[
			(pycurl.SSLCERTTYPE, "PEM"),
			(pycurl.SSLCERT, fspath(tmp_path / STATIC_SHA1)),
		],
	)
	yield SubtestParamsCombined(
		"combined cert and key (P12)",
		"GnuTLS/8.11.0",
		sample_blob_p12,
		[
			(pycurl.SSLCERTTYPE, "P12"),
			(pycurl.SSLCERT, fspath(tmp_path / STATIC_SHA1)),
		],
	)
	yield SubtestParamsCombined(
		"combined cert and key (P12)",
		"GnuTLS/8.10.0",
		sample_blob_p12,
		[
			(pycurl.SSLCERTTYPE, "PEM"),
			(pycurl.SSLCERT, fspath(tmp_path / STATIC_SHA1)),
		],
	)
	yield SubtestParamsSeparate(
		"separate cert (PEM) and key (RSA)",
		"GnuTLS/8.11.0",
		sample_blob_p12,
		sample_file_rsa,
		[
			(pycurl.SSLCERTTYPE, "PEM"),
			(pycurl.SSLCERT, fspath(tmp_path / STATIC_SHA1)),
		],
	)
	yield SubtestParamsCombined(
		"combined cert and key file (P12)",
		"GnuTLS/8.10.0",
		sample_file_p12,
		[
			(pycurl.SSLCERTTYPE, "PEM"),
			(pycurl.SSLCERT, fspath(tmp_path / STATIC_SHA1)),
		],
	)
	yield SubtestParamsCombined(
		"combined cert and key file (PEM)",
		"GnuTLS/8.11.0",
		sample_file_pem,
		[
			(pycurl.SSLCERTTYPE, "PEM"),
			(pycurl.SSLCERT, fspath(sample_file_pem.path)),
		],
	)
	yield SubtestParamsCombined(
		"combined cert and key file (P12)",
		"GnuTLS/8.11.0",
		sample_file_p12,
		[
			(pycurl.SSLCERTTYPE, "P12"),
			(pycurl.SSLCERT, fspath(sample_file_p12.path)),
		],
	)

	### MbedTLS tests
	yield SubtestParamsCombined(
		"combined cert and key (PEM)",
		"MbedTLS/0.0.0",
		sample_blob_pem,
		[
			(pycurl.SSLCERTTYPE, "PEM"),
			(pycurl.SSLCERT_BLOB, sample_blob_pem.to_bytes()),
		],
	)
	yield SubtestParamsCombined(
		"combined cert and key file (PEM)",
		"MbedTLS/0.0.0",
		sample_file_pem,
		[
			(pycurl.SSLCERTTYPE, "PEM"),
			(pycurl.SSLCERT, fspath(sample_file_pem.path)),
		],
	)
	# yield SubtestParamsCombined(
	# 	"combined cert and key (PKCS12)",
	# 	"MbedTLS/0.0.0",
	# 	sample_blob_p12,
	# 	[
	# 		(pycurl.SSLCERTTYPE, "PEM"),
	# 		(
	# 			pycurl.SSLCERT_BLOB,
	# 			AsciiArmored.new(
	# 				certificate=sample_blob_p12.certificate(),
	# 				private_key=sample_blob_p12.private_key(),
	# 			),
	# 		),
	# 	],
	# )
	# yield SubtestParamsCombined(
	# 	"combined cert and key file (PKCS12)",
	# 	"MbedTLS/0.0.0",
	# 	sample_file_p12,
	# 	[
	# 		(pycurl.SSLCERTTYPE, "PEM"),
	# 		(
	# 			pycurl.SSLCERT_BLOB,
	# 			AsciiArmored.new(
	# 				certificate=sample_file_p12.contents.certificate(),
	# 				private_key=sample_file_p12.contents.private_key(),
	# 			),
	# 		),
	# 	],
	# )
	yield SubtestParamsSeparate(
		"separated cert blob (x509) and key blob (RSA)",
		"MbedTLS/0.0.0",
		cert := Certificate(b"xxx"),
		key := RSAPrivateKey(b"xxx"),
		[
			(pycurl.SSLCERTTYPE, "DER"),
			(pycurl.SSLCERT_BLOB, cert.to_bytes()),
			(pycurl.SSLKEYTYPE, "PEM"),
			(pycurl.SSLKEY, fspath(tmp_path / STATIC_SHA1)),
		],
	)
	yield SubtestParamsSeparate(
		"separated cert file (x509) and key file (RSA)",
		"MbedTLS/0.0.0",
		sample_file_x509,
		sample_file_rsa,
		[
			(pycurl.SSLCERTTYPE, "DER"),
			(pycurl.SSLCERT, fspath(sample_file_x509.path)),
			(pycurl.SSLKEYTYPE, "PEM"),
			(pycurl.SSLKEY, fspath(tmp_path / STATIC_SHA1)),
		],
	)
	yield SubtestParamsSeparate(
		"separated cert file (x509) and key file (PEM)",
		"MbedTLS/0.0.0",
		sample_file_x509,
		sample_file_pem,
		[
			(pycurl.SSLCERTTYPE, "DER"),
			(pycurl.SSLCERT, fspath(sample_file_x509.path)),
			(pycurl.SSLKEYTYPE, "PEM"),
			(pycurl.SSLKEY, fspath(sample_file_pem.path)),
		],
	)

	### WolfSSL tests
	yield SubtestParamsCombined(
		"combined cert and key (PEM)",
		"WolfSSL/0.0.0",
		sample_blob_pem,
		[
			(pycurl.SSLCERTTYPE, "PEM"),
			(pycurl.SSLCERT_BLOB, sample_blob_pem.to_bytes()),
		],
	)
	yield SubtestParamsCombined(
		"combined cert and key file (PEM)",
		"WolfSSL/0.0.0",
		sample_file_pem,
		[
			(pycurl.SSLCERTTYPE, "PEM"),
			(pycurl.SSLCERT, fspath(sample_file_pem.path)),
		],
	)
	# yield SubtestParamsCombined(
	# 	"combined cert and key (PKCS12)",
	# 	"WolfSSL/0.0.0",
	# 	sample_blob_p12,
	# 	[
	# 		(pycurl.SSLCERTTYPE, "PEM"),
	# 		(
	# 			pycurl.SSLCERT_BLOB,
	# 			AsciiArmored.new(
	# 				certificate=sample_blob_p12.certificate(),
	# 				private_key=sample_blob_p12.private_key(),
	# 			),
	# 		),
	# 	],
	# )
	# yield SubtestParamsCombined(
	# 	"combined cert and key file (PKCS12)",
	# 	"WolfSSL/0.0.0",
	# 	sample_file_p12,
	# 	[
	# 		(pycurl.SSLCERTTYPE, "PEM"),
	# 		(
	# 			pycurl.SSLCERT_BLOB,
	# 			AsciiArmored.new(
	# 				certificate=sample_file_p12.contents.certificate(),
	# 				private_key=sample_file_p12.contents.private_key(),
	# 			),
	# 		),
	# 	],
	# )
	# yield SubtestParamsSeparate(
	# 	"separated cert blob (DER) and key blob (PKCS12)",
	# 	"WolfSSL/0.0.0",
	# 	cert := Certificate(b"xxx"),
	# 	sample_blob_p12,
	# 	[
	# 		(pycurl.SSLCERTTYPE, "DER"),
	# 		(pycurl.SSLCERT_BLOB, cert.to_bytes()),
	# 		(pycurl.SSLKEYTYPE, "DER"),
	# 		(pycurl.SSLKEY_BLOB, sample_blob_p12.private_key().to_bytes()),
	# 	],
	# )
	yield SubtestParamsSeparate(
		"separated cert blob (x509) and key blob (RSA)",
		"WolfSSL/0.0.0",
		cert := Certificate(b"xxx"),
		key := RSAPrivateKey(b"xxx"),
		[
			(pycurl.SSLCERTTYPE, "DER"),
			(pycurl.SSLCERT_BLOB, cert.to_bytes()),
			(pycurl.SSLKEYTYPE, "DER"),
			(pycurl.SSLKEY_BLOB, key.to_bytes()),
		],
	)
	yield SubtestParamsSeparate(
		"separated cert file (x509) and key file (RSA)",
		"WolfSSL/0.0.0",
		sample_file_x509,
		sample_file_rsa,
		[
			(pycurl.SSLCERTTYPE, "DER"),
			(pycurl.SSLCERT, fspath(sample_file_x509.path)),
			(pycurl.SSLKEYTYPE, "DER"),
			(pycurl.SSLKEY_BLOB, sample_file_rsa.contents.private_key()),
		],
	)
	yield SubtestParamsSeparate(
		"separated cert file (x509) and key file (PEM)",
		"WolfSSL/0.0.0",
		sample_file_x509,
		sample_file_pem,
		[
			(pycurl.SSLCERTTYPE, "DER"),
			(pycurl.SSLCERT, fspath(sample_file_x509.path)),
			(pycurl.SSLKEYTYPE, "PEM"),
			(pycurl.SSLKEY, fspath(sample_file_pem.path)),
		],
	)
